@@ -6,7 +6,7 @@
  *
  ********************************************************************
  *
- * Copyright (C) 1994-2001 Jeff Tranter (tranter@pobox.com)
+ * Copyright (C) 1994-2005 Jeff Tranter (tranter@pobox.com)
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -20,7 +20,7 @@
  *
  * You should have received a copy of the GNU General Public License
  * along with this program; if not, write to the Free Software
- * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
+ * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
  *
  ********************************************************************
  *
@@ -68,6 +68,24 @@
 #include <scsi/scsi.h>
 #include <scsi/sg.h>
 #include <scsi/scsi_ioctl.h>
+#include <sys/time.h>
+
+/* Used by the ToggleTray() function. If ejecting the tray takes this
+ * time or less, the tray was probably already ejected, so we close it
+ * again.
+ */
+#define TRAY_WAS_ALREADY_OPEN_USECS  200000	/* about 0.2 seconds */
+
+
+#define CLOSE(fd) if (close(fd)==-1) { \
+    perror(programName); \
+    exit(1); \
+}
+
+#define FCLOSE(fd) if (fclose(fd)==-1) { \
+    perror(programName); \
+    exit(1); \
+}
 
 /* Global Variables */
 const char *version = VERSION; /* program version */
@@ -81,9 +99,11 @@ int q_option = 0;
 int r_option = 0;
 int s_option = 0;
 int t_option = 0;
+int T_option = 0;
 int v_option = 0;
 int x_option = 0;
 int p_option = 0;
+int m_option = 0;
 int a_arg = 0;
 int c_arg = 0;
 int x_arg = 0;
@@ -120,6 +140,7 @@ static void usage()
 "  eject [-vn] -a on|off|1|0 [<name>]	-- turn auto-eject feature on or off\n"
 "  eject [-vn] -c <slot> [<name>]	-- switch discs on a CD-ROM changer\n"
 "  eject [-vn] -t [<name>]		-- close tray\n"
+"  eject [-vn] -T [<name>]		-- toggle tray\n"
 "  eject [-vn] -x <speed> [<name>]	-- set CD-ROM max speed\n"
 "Options:\n"
 "  -v\t-- enable verbose output\n"
@@ -128,7 +149,9 @@ static void usage()
 "  -s\t-- eject SCSI device\n"
 "  -f\t-- eject floppy\n"
 "  -q\t-- eject tape\n"
-"  -p\t-- use /proc/mounts instead of /etc/mtab\n")
+"  -p\t-- use /proc/mounts instead of /etc/mtab\n"
+"  -m\t-- do not unmount device even if it is mounted\n"
+)
 , version);
 #ifdef GETOPTLONG
 	fprintf(stderr,_(
@@ -137,7 +160,7 @@ static void usage()
 "  -a --auto   -c --changerslot  -t --trayclose  -x --cdspeed\n"
 "  -r --cdrom  -s --scsi	 -f --floppy\n"
 "  -q --tape   -n --noop	 -V --version\n"
-"  -p --proc\n"));
+"  -p --proc   -m --no-unmount -T --traytoggle\n"));
 #endif /* GETOPTLONG */
 	fprintf(stderr,_(
 "Parameter <name> can be a device file or a mount point.\n"
@@ -151,7 +174,7 @@ static void usage()
 /* Handle command line options. */
 static void parse_args(int argc, char **argv, char **device)
 {
-	const char *flags = "a:c:x:dfhnqrstvVp";
+	const char *flags = "a:c:x:dfhnqrstTvVpm";
 #ifdef GETOPTLONG
 	static struct option long_options[] =
 	{
@@ -161,6 +184,7 @@ static void parse_args(int argc, char **argv, char **device)
 		{"auto",	required_argument, NULL, 'a'},
 		{"changerslot", required_argument, NULL, 'c'},
 		{"trayclose",	no_argument,	   NULL, 't'},
+		{"traytoggle",	no_argument,	   NULL, 'T'},
 		{"cdspeed",	required_argument, NULL, 'x'},
 		{"noop",	no_argument,	   NULL, 'n'},
 		{"cdrom",	no_argument,	   NULL, 'r'},
@@ -169,6 +193,7 @@ static void parse_args(int argc, char **argv, char **device)
 		{"tape",	no_argument,	   NULL, 'q'},
 		{"version",	no_argument,	   NULL, 'V'},
 		{"proc",	no_argument,	   NULL, 'p'},
+		{"no-unmount",	no_argument,	   NULL, 'm'},
 		{0, 0, 0, 0}
 	};
 	int option_index;
@@ -231,6 +256,9 @@ static void parse_args(int argc, char **argv, char **device)
 			  usage();
 			  exit(0);
 			  break;
+		  case 'm':
+			  m_option = 1;
+			  break;
 		  case 'n':
 			  n_option = 1;
 			  break;
@@ -249,6 +277,9 @@ static void parse_args(int argc, char **argv, char **device)
 		  case 't':
 			  t_option = 1;
 			  break;
+		  case 'T':
+			  T_option = 1;
+			  break;
 		  case 'v':
 			  v_option = 1;
 			  break;
@@ -294,6 +325,7 @@ static int FileExists(const char *name)
  *
  * foo (if foo starts with '.' or '/')
  * /dev/foo
+ * /media/foo
  * /mnt/foo
  * /dev/cdroms/foo
  * /dev/cdroms/foo0
@@ -309,7 +341,10 @@ static char *FindDevice(const char *name)
 	char *buf;
 
 	buf = (char *) malloc(strlen(name)+14); /* to allow for "/dev/cdroms/ + "0" + null */
-
+	if (buf==NULL) {
+		fprintf(stderr, _("%s: could not allocate memory\n"), programName);
+		exit(1);
+	}
 	if ((name[0] == '.') || (name[0] == '/')) {
 		strcpy(buf, name);
 		if (FileExists(buf))
@@ -321,6 +356,11 @@ static char *FindDevice(const char *name)
 	if (FileExists(buf))
 		return buf;
 
+	strcpy(buf, "/media/");
+	strcat(buf, name);
+	if (FileExists(buf))
+		return buf;
+	
 	strcpy(buf, "/mnt/");
 	strcat(buf, name);
 	if (FileExists(buf))
@@ -417,6 +457,53 @@ static void CloseTray(int fd)
 #endif
 }
 
+/*
+ * Toggle tray.
+ *
+ * Written by Benjamin Schwenk <benjaminschwenk@yahoo.de> and
+ * Sybren Stuvel <sybren@thirdtower.com>
+ *
+ * Not supported by older kernels because it might use
+ * CloseTray().
+ *
+ */
+static void ToggleTray(int fd)
+{
+	struct timeval time_start, time_stop;
+	int time_elapsed;
+
+#ifdef CDROMCLOSETRAY
+
+	/* Try to open the CDROM tray and measure the time therefor
+	 * needed.  In my experience the function needs less than 0.05
+	 * seconds if the tray was already open, and at least 1.5 seconds
+	 * if it was closed.  */
+	gettimeofday(&time_start, NULL);
+	
+	/* Send the CDROMEJECT command to the device. */
+	if (ioctl(fd, CDROMEJECT, 0) < 0) {
+		perror("ioctl");
+		exit(1);
+	}
+
+	/* Get the second timestamp, to measure the time needed to open
+	 * the tray.  */
+	gettimeofday(&time_stop, NULL);
+
+	time_elapsed = (time_stop.tv_sec * 1000000 + time_stop.tv_usec) -
+		(time_start.tv_sec * 1000000 + time_start.tv_usec);
+
+	/* If the tray "opened" too fast, we can be nearly sure, that it
+	 * was already open. In this case, close it now. Else the tray was
+	 * closed before. This would mean that we are done.  */
+	if (time_elapsed < TRAY_WAS_ALREADY_OPEN_USECS)
+		CloseTray(fd);
+
+#else
+    fprintf(stderr, _("%s: CD-ROM tray toggle command not supported by this kernel\n"), programName);
+#endif
+	
+}
 
 /*
  * Select Speed of CD-ROM drive.
@@ -627,7 +714,7 @@ static int MountedDevice(const char *name, char **mountName, char **deviceName)
 			GetMajorMinor(s1, &mtabmaj, &mtabmin);
 			if (((strcmp(s1, name) == 0) || (strcmp(s2, name) == 0)) ||
 				((maj != -1) && (maj == mtabmaj) && (min == mtabmin))) {
-				fclose(fp);
+				FCLOSE(fp);
 				*deviceName = strdup(s1);
 				*mountName = strdup(s2);
 				return 1;
@@ -636,7 +723,7 @@ static int MountedDevice(const char *name, char **mountName, char **deviceName)
 	}
 	*deviceName = 0;
 	*mountName = 0;
-	fclose(fp);
+	FCLOSE(fp);
 	return 0;
 }
 
@@ -656,20 +743,28 @@ static int MountableDevice(const char *name, char **mountName, char **deviceName
 
 	fp = fopen("/etc/fstab", "r");
 	if (fp == NULL) {
-		fprintf(stderr, _("%s: unable to open /etc/fstab: %s\n"), programName, strerror(errno));
-		exit(1);
+/*
+ * /etc/fstab may be unreadable in some situations due to passwords in the
+ * file.
+ */
+/*		fprintf(stderr, _("%s: unable to open /etc/fstab: %s\n"), programName, strerror(errno));
+		exit(1);*/
+		if (v_option) {
+			printf( _("%s: unable to open /etc/fstab: %s\n"), programName, strerror(errno));
+		}
+		return -1;
 	}
 
 	while (fgets(line, sizeof(line), fp) != 0) {
 		rc = sscanf(line, "%1023s %1023s", s1, s2);
 		if (rc >= 2 && s1[0] != '#' && strcmp(s2, name) == 0) {
-			fclose(fp);
+			FCLOSE(fp);
 			*deviceName = strdup(s1);
 			*mountName = strdup(s2);
 			return 1;
 		}
 	}
-	fclose(fp);
+	FCLOSE(fp);
 	return 0;
 }
 
@@ -687,7 +782,10 @@ static void UnmountDevices(const char *pattern)
 	char line[1024];
 	int status;
 
-	regcomp(&preg, pattern, REG_EXTENDED);
+	if (regcomp(&preg, pattern, REG_EXTENDED)!=0) {
+		perror(programName);
+		exit(1);
+	}
 
 	fp = fopen((p_option ? "/proc/mounts" : "/etc/mtab"), "r");
 	if (fp == NULL)
@@ -708,7 +806,7 @@ static void UnmountDevices(const char *pattern)
 			}
 		}
 	}
-	fclose(fp);
+	FCLOSE(fp);
 }
 
 
@@ -890,7 +988,8 @@ int main(int argc, char **argv)
 	/* if not currently mounted, see if it is a possible mount point */
 	if (!mounted) {
 		mountable = MountableDevice(fullName, &mountName, &deviceName);
-		if (v_option) {
+		/* if return value -1 then fstab could not be read */
+		if (v_option && mountable >= 0) {
 			if (mountable)
 				printf(_("%s: `%s' can be mounted at `%s'\n"), programName, deviceName, mountName);
 			else
@@ -929,11 +1028,21 @@ int main(int argc, char **argv)
 		exit(0);
 	}
 
+	/* handle -T option */
+	if (T_option) {
+		if (v_option)
+			printf(_("%s: toggling tray\n"), programName);
+		fd = OpenDevice(deviceName);
+		ToggleTray(fd);
+		HandleXOption(deviceName);
+		exit(0);
+	}
+
 	/* handle -x option only */
 	if (!c_option) HandleXOption(deviceName);
 
 	/* unmount device if mounted */
-	if (mounted) {
+	if ((m_option != 1) && mounted) {
 		if (v_option)
 			printf(_("%s: unmounting `%s'\n"), programName, deviceName);
 		Unmount(deviceName);
@@ -942,7 +1051,7 @@ int main(int argc, char **argv)
 	/* if it is a multipartition device, unmount any other partitions on
 	   the device */
 	pattern = MultiplePartitions(deviceName);
-	if (pattern != 0)
+	if ((m_option != 1) && (pattern != 0))
 		UnmountDevices(pattern);
 
 	/* handle -c option */
@@ -1018,7 +1127,7 @@ int main(int argc, char **argv)
 	}
 
 	/* cleanup */
-	close(fd);
+	CLOSE(fd);
 	free(device);
 	free(deviceName);
 	free(fullName);
